@@ -2,7 +2,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2005-2015 The OpenLDAP Foundation.
+ * Copyright 2005-2017 The OpenLDAP Foundation.
  * Portions copyright 2004-2005 Symas Corporation.
  * All rights reserved.
  *
@@ -578,7 +578,7 @@ log_age_unparse( int age, struct berval *agebv, size_t size )
 	agebv->bv_len = ptr - agebv->bv_val;
 }
 
-static slap_callback nullsc = { NULL, NULL, NULL, NULL };
+static slap_callback nullsc;
 
 #define PURGE_INCREMENT	100
 
@@ -637,7 +637,7 @@ accesslog_purge( void *ctx, void *arg )
 	OperationBuffer opbuf;
 	Operation *op;
 	SlapReply rs = {REP_RESULT};
-	slap_callback cb = { NULL, log_old_lookup, NULL, NULL };
+	slap_callback cb = { NULL, log_old_lookup, NULL, NULL, NULL };
 	Filter f;
 	AttributeAssertion ava = ATTRIBUTEASSERTION_INIT;
 	purge_data pd = {0};
@@ -701,6 +701,7 @@ accesslog_purge( void *ctx, void *arg )
 			}
 			ch_free( pd.ndn[i].bv_val );
 			ch_free( pd.dn[i].bv_val );
+			ldap_pvt_thread_pool_pausecheck( &connection_pool );
 		}
 		ch_free( pd.ndn );
 		ch_free( pd.dn );
@@ -1490,9 +1491,17 @@ static int accesslog_response(Operation *op, SlapReply *rs) {
 	if ( lo->mask & LOG_OP_WRITES ) {
 		slap_callback *cb;
 
-		/* These internal ops are not logged */
-		if ( op->o_dont_replicate && op->orm_no_opattrs )
+		/* Most internal ops are not logged */
+		if ( op->o_dont_replicate) {
+			/* Let contextCSN updates from syncrepl thru; the underlying
+			 * syncprov needs to see them. Skip others.
+			 */
+			if (( op->o_tag != LDAP_REQ_MODIFY ||
+				op->orm_modlist->sml_op != LDAP_MOD_REPLACE ||
+				op->orm_modlist->sml_desc != slap_schema.si_ad_contextCSN ) &&
+				op->orm_no_opattrs )
 			return SLAP_CB_CONTINUE;
+		}
 
 		ldap_pvt_thread_mutex_lock( &li->li_log_mutex );
 		old = li->li_old;
@@ -1577,7 +1586,7 @@ static int accesslog_response(Operation *op, SlapReply *rs) {
 
 	case LOG_EN_MODRDN:
 	case LOG_EN_MODIFY:
-		/* count all the mods */
+		/* count all the mods + attributes (ITS#6545) */
 		i = 0;
 		for ( m = op->orm_modlist; m; m = m->sml_next ) {
 			if ( m->sml_values ) {
@@ -1587,6 +1596,9 @@ static int accesslog_response(Operation *op, SlapReply *rs) {
 			{
 				i++;
 			}
+			if ( m->sml_next && m->sml_desc == m->sml_next->sml_desc ) {
+				i++;
+			}
 		}
 		vals = ch_malloc( (i+1) * sizeof( struct berval ));
 		i = 0;
@@ -1657,6 +1669,12 @@ static int accesslog_response(Operation *op, SlapReply *rs) {
 				*ptr = '\0';
 				i++;
 			}
+			/* ITS#6545: when the same attribute is edited multiple times,
+			 * record the transition */
+			if ( m->sml_next && m->sml_desc == m->sml_next->sml_desc ) {
+				ber_str2bv( ":", STRLENOF(":"), 1, &vals[i] );
+				i++;
+			}
 		}
 
 		if ( i > 0 ) {
@@ -1831,10 +1849,29 @@ static int accesslog_response(Operation *op, SlapReply *rs) {
 	op2.o_req_ndn = e->e_nname;
 	op2.ora_e = e;
 	op2.o_callback = &nullsc;
+	op2.o_csn = op->o_csn;
+	/* contextCSN updates may still reach here */
+	op2.o_dont_replicate = op->o_dont_replicate;
 
 	if (( lo->mask & LOG_OP_WRITES ) && !BER_BVISEMPTY( &op->o_csn )) {
-		slap_queue_csn( &op2, &op->o_csn );
-		do_graduate = 1;
+		struct berval maxcsn;
+		char cbuf[LDAP_PVT_CSNSTR_BUFSIZE];
+		int foundit;
+		cbuf[0] = '\0';
+		maxcsn.bv_val = cbuf;
+		maxcsn.bv_len = sizeof(cbuf);
+		/* If there was a commit CSN on the main DB,
+		 * we must propagate it to the log DB for its
+		 * own syncprov. Otherwise, don't generate one.
+		 */
+		slap_get_commit_csn( op, &maxcsn, &foundit );
+		if ( !BER_BVISEMPTY( &maxcsn ) ) {
+			slap_queue_csn( &op2, &op->o_csn );
+			do_graduate = 1;
+		} else {
+			attr_merge_normalize_one( e, slap_schema.si_ad_entryCSN,
+				&op->o_csn, op->o_tmpmemctx );
+		}
 	}
 
 	op2.o_bd->be_add( &op2, &rs2 );
@@ -1842,8 +1879,6 @@ static int accesslog_response(Operation *op, SlapReply *rs) {
 	e = NULL;
 	if ( do_graduate ) {
 		slap_graduate_commit_csn( &op2 );
-		if ( op2.o_csn.bv_val )
-			op->o_tmpfree( op2.o_csn.bv_val, op->o_tmpmemctx );
 	}
 
 done:
@@ -1922,8 +1957,18 @@ accesslog_op_mod( Operation *op, SlapReply *rs )
 	int doit = 0;
 
 	/* These internal ops are not logged */
-	if ( op->o_dont_replicate && op->orm_no_opattrs )
+	if ( op->o_dont_replicate ) {
+		/* Let contextCSN updates from syncrepl thru; the underlying
+		 * syncprov needs to see them. Skip others.
+		 */
+		if (( op->o_tag != LDAP_REQ_MODIFY ||
+			op->orm_modlist->sml_op != LDAP_MOD_REPLACE ||
+			op->orm_modlist->sml_desc != slap_schema.si_ad_contextCSN ) &&
+			op->orm_no_opattrs )
 		return SLAP_CB_CONTINUE;
+		/* give this a unique timestamp */
+		op->o_tincr++;
+	}
 
 	logop = accesslog_op2logop( op );
 	lo = logops+logop+EN_OFFSET;
@@ -1940,11 +1985,9 @@ accesslog_op_mod( Operation *op, SlapReply *rs )
 	}
 			
 	if ( doit ) {
-		slap_callback *cb = op->o_tmpalloc( sizeof( slap_callback ), op->o_tmpmemctx ), *cb2;
+		slap_callback *cb = op->o_tmpcalloc( 1, sizeof( slap_callback ), op->o_tmpmemctx ), *cb2;
 		cb->sc_cleanup = accesslog_mod_cleanup;
-		cb->sc_response = NULL;
 		cb->sc_private = on;
-		cb->sc_next = NULL;
 		for ( cb2 = op->o_callback; cb2->sc_next; cb2 = cb2->sc_next );
 		cb2->sc_next = cb;
 
